'''
Functions for parsing line-by-line.
'''
from typing import List, Iterator, Tuple, Type
import re
import enum

import attr


@attr.s
class RawSourceLine:
    '''A line of source code'''
    number = attr.ib()  # int
    type = attr.ib()  # RawLineType
    raw = attr.ib()  # str

    def as_chunk_line(self) -> 'ChunkLine':
        chunk_line_type = ChunkLineType.for_source_line(self)
        return ChunkLine(self.number, chunk_line_type, self.raw)

    def as_comment_line(self) -> 'ChunkLine':
        chunk_line_type = ChunkLineType.Comment
        return ChunkLine(self.number, chunk_line_type, self.raw)


class RawLineType(enum.Enum):
    PortModuleStart = re.compile(r'^port\s+module')
    PortFunctionStart = re.compile(r'^port\s+[a-z]')
    MultilineCommentInOne = re.compile(r'^\s*\{-.*-\}\s*$')
    MultilineCommentStart = re.compile(r'^\s*\{-')
    MultilineCommentEnd = re.compile(r'-\}\s*$')
    InlineComment = re.compile(r'^\s*--')
    TopLevelThing = re.compile(r'^\S')
    IndentedThing = re.compile(r'^\s+\S')
    EmptyLine = re.compile(r'^\s*$')


def _iter_source_lines(source: str) -> Iterator[RawSourceLine]:
    # The True argument keeps newlines as part of the split results
    for i, line in enumerate(source.splitlines(True)):
        for line_type in RawLineType:
            if line_type.value and line_type.value.search(line):
                yield RawSourceLine(i + 1, line_type, line)
                break
        else:
            raise Exception('Could not determine what kind of source line this is at line {}: {}'.format(
                i + 1, line))


@attr.s
class ChunkLine:
    '''A line of source code in a chunk'''
    number = attr.ib()  # int
    type = attr.ib()  # ChunkLineType
    raw = attr.ib()  # str

    def is_source(self) -> bool:
        return self.type == ChunkLineType.Source


class ChunkLineType(enum.Enum):
    Source = 1
    Comment = 2
    Empty = 3

    @classmethod
    def for_source_line(cls, source_line: RawSourceLine) -> 'ChunkLineType':
        if source_line.type in (RawLineType.PortModuleStart,
                                RawLineType.PortFunctionStart,
                                RawLineType.TopLevelThing,
                                RawLineType.IndentedThing):
            return cls.Source

        elif source_line.type in (RawLineType.MultilineCommentInOne,
                                  RawLineType.MultilineCommentStart,
                                  RawLineType.MultilineCommentEnd,
                                  RawLineType.InlineComment):
            return cls.Comment

        elif source_line.type in (RawLineType.EmptyLine,):
            return cls.Empty

        raise ValueError('Unknown RawLineType: {}. This is likely a bug'.format(source_line.type))


@attr.s
class Chunk:
    '''A semi-logical group of continuous SourceLines'''
    type = attr.ib()  # ChunkType
    lines = attr.ib()  # List[ChunkLine]

    def is_port_function(self) -> bool:
        return self.type == ChunkType.PortFunctionDeclaration

    def is_port_module(self) -> bool:
        return self.type == ChunkType.PortModuleDeclaration

    def raw_lines(self) -> List[str]:
        return [line.raw for line in self.lines]

    def non_comment_raw_lines(self) -> Iterator[str]:
        if self.type == ChunkType.MultilineComment:
            yield from ()
        for line in self.lines:
            if line.is_source():
                yield line.raw


class ChunkType(enum.Enum):
    PortModuleDeclaration = 1
    PortFunctionDeclaration = 2
    MultilineComment = 3
    Other = 4


@attr.s
class ChunkStore:
    '''Datastructure to hold Chunks during parsing'''
    current_chunk = attr.ib(default=None)  # Chunk
    ready_to_yield = attr.ib(default=None)  # Chunk

    def add(self, line: ChunkLine):
        '''Add a SourceLine to the currently active Chunk.'''
        if not self.current_chunk:
            self.current_chunk = Chunk(ChunkType.Other, [])
        self.current_chunk.lines.append(line)

    def add_and_pop(self, line: ChunkLine):
        '''Add a SourceLine to the currently active Chunk and mark as done.'''
        self.add(line)
        self.pop()

    def pop_and_add(self, chunk_type: ChunkType, line: ChunkLine):
        '''Mark the currently active Chunk as done and then
        start a new Chunk with the given type and starting line.'''
        self.pop()
        self.current_chunk = Chunk(chunk_type, [line])

    def pop(self):
        '''Mark the currently active Chunk as done.'''
        if self.current_chunk:
            self.ready_to_yield = self.current_chunk
            self.current_chunk = None

    def take(self):
        '''Move the chunk marked as done off of this store.'''
        chunk = self.ready_to_yield
        self.ready_to_yield = None
        return chunk


class ParsingState:
    '''What the parser is currently doing.'''


ParsingStateStack = List[Type[ParsingState]]


class NothingSpecial(ParsingState):
    @classmethod
    def on_source_line(cls,
                       line: RawSourceLine,
                       state_stack: ParsingStateStack,
                       store: ChunkStore) -> Tuple[ParsingStateStack, ChunkStore]:
        if line.type == RawLineType.PortModuleStart:
            # this line is the start of the next chunk, which is a port module declaration
            store.pop_and_add(ChunkType.PortModuleDeclaration, line.as_chunk_line())
            return ([DefiningPortModule], store)

        elif line.type == RawLineType.PortFunctionStart:
            # this line is the start of the next chunk, which is a port function
            store.pop_and_add(ChunkType.PortFunctionDeclaration, line.as_chunk_line())
            return ([DefiningPortFunction], store)

        elif line.type == RawLineType.MultilineCommentStart:
            # this line is the start of the next chunk, which is a multiline comment
            store.pop_and_add(ChunkType.MultilineComment, line.as_chunk_line())
            return ([DefiningMultilineComment], store)

        else:
            # this line is nothing special; add to the current chunk
            store.add(line.as_chunk_line())
            return (state_stack, store)


class DefiningPortModule(ParsingState):
    @classmethod
    def on_source_line(cls,
                       line: RawSourceLine,
                       state_stack: ParsingStateStack,
                       store: ChunkStore) -> Tuple[ParsingStateStack, ChunkStore]:
        if line.type == RawLineType.MultilineCommentStart:
            # this line is the start of a comment inside the port module declaration we're defining
            store.add(line.as_chunk_line())
            state_stack.append(DefiningMultilineComment)
            return (state_stack, store)

        elif line.type == RawLineType.PortFunctionStart:
            # this line is the start of the next chunk, which is a port function
            store.pop_and_add(ChunkType.PortFunctionDeclaration, line.as_chunk_line())
            return ([DefiningPortFunction], store)

        elif line.type == RawLineType.TopLevelThing:
            # this line is the start of the next chunk
            store.pop_and_add(ChunkType.Other, line.as_chunk_line())
            return ([NothingSpecial], store)

        else:
            # this line is part of the port module declaration we're defining.
            # note: this adds inline comments to the port declaration lines.
            store.add(line.as_chunk_line())
            return (state_stack, store)


class DefiningPortFunction(ParsingState):
    @classmethod
    def on_source_line(cls,
                       line: RawSourceLine,
                       state_stack: ParsingStateStack,
                       store: ChunkStore) -> Tuple[ParsingStateStack, ChunkStore]:
        if line.type == RawLineType.MultilineCommentStart:
            # this line is the start of a comment inside the port function we're defining
            store.add(line.as_chunk_line())
            state_stack.append(DefiningMultilineComment)
            return (state_stack, store)

        elif line.type == RawLineType.PortFunctionStart:
            # this line is the start of the next chunk, which is a port function
            store.pop_and_add(ChunkType.PortFunctionDeclaration, line.as_chunk_line())
            return ([DefiningPortFunction], store)

        elif line.type == RawLineType.TopLevelThing:
            # this line is the start of the next chunk
            store.pop_and_add(ChunkType.Other, line.as_chunk_line())
            state_stack = [NothingSpecial]
            return (state_stack, store)

        else:
            # this line is part of the port function we're defining.
            # note: this adds inline comments to the port declaration lines.
            store.add(line.as_chunk_line())
            return (state_stack, store)


class DefiningMultilineComment(ParsingState):
    @classmethod
    def on_source_line(cls,
                       line: RawSourceLine,
                       state_stack: ParsingStateStack,
                       store: ChunkStore) -> Tuple[ParsingStateStack, ChunkStore]:
        if line.type == RawLineType.MultilineCommentEnd:
            if len(state_stack) > 1:
                # we were defining a multiline comment inside of a port declaration.
                # add this line to the port declaration and pop the current state.
                store.add(line.as_chunk_line())
                state_stack = state_stack[:-1]
            else:
                # we were defining a top-level multiline comment.
                # add this very last line to that comment and mark it as done.
                store.add_and_pop(line.as_chunk_line())
                state_stack = [NothingSpecial]
            return (state_stack, store)

        else:
            store.add(line.as_comment_line())
            return (state_stack, store)


def iter_line_chunks(source: str) -> Iterator[Chunk]:
    state_stack = [NothingSpecial]
    store = ChunkStore()
    for line in _iter_source_lines(source):
        state_stack, store = state_stack[-1].on_source_line(line, state_stack, store)
        if store.ready_to_yield:
            yield store.take()
    if store.ready_to_yield:
        yield store.take()
    store.pop()
    if store.ready_to_yield:
        yield store.take()


if __name__ == '__main__':
    import sys
    with open(sys.argv[1]) as f:
        source = f.read()
    for chunk in iter_line_chunks(source):
        for line in chunk.lines:
            raw = line[:-1] if line and line.raw[-1] == '\n' else line
            print(line.number, chunk.type.name, line.type.name, raw)



import sys
import os
import os.path
import shutil
from pathlib import Path

import click
from doit.doit_cmd import DoitMain
from doit.cmd_base import ModuleTaskLoader

from elm_doc.tasks import build_task_creators
from elm_doc.elm_project import ProjectConfig


class DoitException(click.ClickException):
    def __init__(self, message, exit_code):
        click.ClickException.__init__(self, message)
        self.exit_code = exit_code


def validate_mount_at(ctx, param, value):
    if len(value) == 0 or value.startswith('/'):
        return value
    else:
        raise click.BadParameter('should be an absolute path, like /{}'.format(value))


def validate_elm_path(ctx, param, value):
    if value is None:
        return value

    realpath = os.path.realpath(value)
    if not os.path.isfile(realpath):
        realpath = shutil.which(value)

    if realpath is None or not os.path.isfile(realpath):
        raise click.BadParameter('{} not found'.format(value))


def _resolve_path(path: str) -> Path:
    # not using Path.resolve() for now because we don't expect strict
    # existence checking. maybe we should.
    return Path(os.path.normpath(os.path.abspath(path)))


class LazyOutfile:
    def write(self, *args, **kwargs):
        sys.stdout.write(*args, **kwargs)


@click.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.option('--output', '-o',
              metavar='dir')
@click.option('--build-dir',
              metavar='dir',
              help=('temporary build directory. source files will be copied here. '
                    'default: <project_path>/.elm-doc/'))
@click.option('--elm-path',
              metavar='path/to/elm',
              callback=validate_elm_path,
              help=('specify which elm binary to use. if not specified, '
                    'elm will be installed afresh in a temporary directory'))
@click.option('--mount-at',
              metavar='/path',
              default='',
              callback=validate_mount_at,
              help='url path at which the docs will be served')
@click.option('--exclude', '-x',
              metavar='module1,module2.*',
              help='comma-separated fnmatch pattern of modules to exclude from the list of included modules')
@click.option('--fake-user',
              metavar='GitHub user name',
              default='user',
              help='User name to use as part of the package name of the project as listed in the generated docs')
@click.option('--fake-project',
              metavar='GitHub repo name',
              default='project',
              help='Project name to use as part of the package name of the project as listed in the generated docs')
@click.option('--fake-version',
              metavar='Semver version string',
              default='1.0.0',
              help='Version of the project as listed in the generated docs')
@click.option('--fake-summary',
              metavar='Project summary',
              default='project summary',
              help='Summary of the project as listed in the generated docs')
@click.option('--fake-license',
              metavar='OSI-approved SPDX liense',
              required=True,
              help='License of the project to tell the Elm compiler  when generating docs')
@click.option('--force-exclusion/--no-force-exclusion',
              default=False,
              help=('force excluding modules specified by --exclude even if '
                    'they are explicitly specified as include_paths'))
@click.option('--validate/--no-validate',
              default=False,
              help='validate all doc comments are in place without generating docs')
@click.option('--doit-args',
              help='options to pass to doit.doit_cmd.DoitMain.run')
@click.argument('project_path')
@click.argument('include_paths', nargs=-1)
def main(
        output,
        build_dir,
        elm_path,
        mount_at,
        exclude,
        force_exclusion,
        fake_user,
        fake_project,
        fake_version,
        fake_summary,
        fake_license,
        validate,
        doit_args,
        project_path,
        include_paths):
    """Generate static documentation for your Elm project"""

    if not validate and output is None:
        raise click.BadParameter('please specify --output directory')

    resolved_include_paths = [_resolve_path(path) for path in include_paths]
    exclude_modules = exclude.split(',') if exclude else []
    project_config = ProjectConfig(
        include_paths=resolved_include_paths,
        exclude_modules=exclude_modules,
        force_exclusion=force_exclusion,
        fake_user=fake_user,
        fake_project=fake_project,
        fake_version=fake_version,
        fake_summary=fake_summary,
        fake_license=fake_license,
    )

    task_creators = build_task_creators(
        _resolve_path(project_path),
        project_config,
        _resolve_path(output) if output is not None else None,
        build_path=_resolve_path(build_dir) if build_dir is not None else None,
        elm_path=_resolve_path(elm_path) if elm_path is not None else None,
        mount_point=mount_at,
        validate=validate)

    extra_config = {'GLOBAL': {'outfile': LazyOutfile()}}
    result = DoitMain(ModuleTaskLoader(task_creators), extra_config=extra_config).run(
        doit_args.split(' ') if doit_args else [])
    if result is not None and result > 0:
        raise DoitException('see output above', result)


if __name__ == '__main__':
    main()



'''
'''
from typing import Optional
from pathlib import Path

from doit import create_after

from elm_doc import elm_project
from elm_doc import project_tasks
from elm_doc import package_tasks
from elm_doc import asset_tasks
from elm_doc import catalog_tasks


def build_task_creators(
        project_path: Path,
        project_config: elm_project.ProjectConfig,
        output_path: Optional[Path] = None,
        build_path: Optional[Path] = None,
        elm_path: Optional[Path] = None,
        mount_point: str = '',
        validate: bool = False):
    # todo: gracefully handle missing elm-package.json
    project = elm_project.from_path(project_path)
    project.add_direct_dependencies(
        catalog_tasks.missing_popular_packages(list(project.all_dependency_names())))

    if build_path is None:
        build_path = project_path / '.elm-doc'

    task_creators = {}

    task_creators['task_main_project'] = build_main_project_task_creator(
        project,
        project_config,
        output_path,
        build_path,
        elm_path,
        mount_point,
        validate,
    )

    if validate:
        return task_creators

    task_creators['task_dependencies'] = build_dependencies_task_creator(
        project,
        project_config,
        output_path,
        mount_point,
    )
    task_creators['task_assets'] = build_assets_task_creator(output_path)

    return task_creators


def build_main_project_task_creator(
        project: elm_project.ElmProject,
        project_config: elm_project.ProjectConfig,
        output_path: Optional[Path] = None,
        build_path: Optional[Path] = None,
        elm_path: Optional[Path] = None,
        mount_point: str = '',
        validate: bool = False):
    def task_main_project():
        for task in project_tasks.create_main_project_tasks(
                project,
                project_config,
                output_path,
                build_path=build_path,
                elm_path=elm_path,
                mount_point=mount_point,
                validate=validate):
            yield task
    return task_main_project


def build_dependencies_task_creator(
        project: elm_project.ElmProject,
        project_config: elm_project.ProjectConfig,
        output_path: Optional[Path] = None,
        mount_point: str = ''):
    @create_after(executed='build_docs_json', creates=[
        # package tasks
        'dep_copy_docs_json', 'dep_top_page', 'dep_readme',
        'dep_releases', 'dep_latest_link', 'dep_module_page',
        # catalog tasks
        'index', 'search_json', 'help',
    ])
    def task_dependencies():
        deps = list(project.iter_dependencies())
        deps.sort(key=lambda dep: dep.name)
        all_packages = [project.as_package(project_config).without_license()] + deps

        for package in deps:
            for task in package_tasks.create_dependency_tasks(
                    output_path, package, mount_point):
                yield task

        for task in catalog_tasks.create_catalog_tasks(all_packages, output_path, mount_point=mount_point):
            yield task
    return task_dependencies


def build_assets_task_creator(output_path: Optional[Path] = None):
    def task_assets():
        yield {
            'basename': 'assets',
            'actions': [(asset_tasks.extract_assets, (output_path,))],
            'targets': [output_path / path for path in asset_tasks.bundled_assets],
            'file_dep': [asset_tasks.tarball]
        }
    return task_assets



from typing import List, Iterator, Tuple
from pathlib import Path
import json

import attr
import requests
from retrying import retry
from doit.tools import config_changed

from elm_doc.elm_project import ElmPackage, ExactVersion
from elm_doc import asset_tasks
from elm_doc import page_tasks


popular_packages = [
    'elm/core',
    'elm/html',
    'elm/json',
    'elm/browser',
    'elm/url',
    'elm/http',
]


def missing_popular_packages(package_names: List[str]) -> Iterator[Tuple[str, ExactVersion]]:
    missing_names = set(popular_packages) - set(package_names)
    for missing_name in missing_names:
        latest_version = _fetch_latest_version(missing_name)
        yield (missing_name, latest_version)


@retry(
    retry_on_exception=lambda e: isinstance(e, requests.RequestException),
    wait_exponential_multiplier=1000,  # Wait 2^x * 1000 milliseconds between each retry,
    wait_exponential_max=30 * 1000,  # up to 30 seconds, then 30 seconds afterwards
    stop_max_attempt_number=10)
def _fetch_latest_version(package_name: str) -> ExactVersion:
    releases_url = 'https://package.elm-lang.org/packages/{}/releases.json'.format(package_name)
    releases = requests.get(releases_url).json()
    return sorted(releases.keys())[-1]


@attr.s
class SearchEntry:
    name = attr.ib()  # str
    summary = attr.ib()  # str
    license = attr.ib()  # str
    versions = attr.ib()  # List[ExactVersion]

    @classmethod
    def from_package(cls, package: ElmPackage) -> 'SearchEntry':
        return SearchEntry(
            name=package.name,
            summary=package.summary,
            license=package.license,
            versions=[package.version],
        )


def write_search_json(entries: List[SearchEntry], output_path: Path):
    with open(str(output_path), 'w') as f:
        json.dump([attr.asdict(entry) for entry in entries], f)


def create_catalog_tasks(packages: List[ElmPackage], output_path: Path, mount_point: str = ''):
    page_flags = {'mount_point': mount_point}

    # index
    index_path = output_path / 'index.html'
    yield {
        'basename': 'index',
        'actions': [(page_tasks.write_page, (index_path,), page_flags)],
        'targets': [index_path],
        'uptodate': [config_changed(page_flags)],
    }

    # search.json
    search_json_path = output_path / 'search.json'
    search_entries = list(map(SearchEntry.from_package, packages))
    yield {
        'basename': 'search_json',
        'actions': [(write_search_json, (search_entries, search_json_path))],
        'targets': [search_json_path],
        'uptodate': [config_changed({'entries': [attr.asdict(entry) for entry in search_entries]})],
    }

    # help pages
    for help_file in asset_tasks.bundled_helps:
        url_path = Path(help_file).relative_to('assets').with_suffix('')
        help_output_path = (output_path / url_path)
        yield {
            'basename': 'help',
            'name': url_path,
            'actions': [(page_tasks.write_page, (help_output_path,), page_flags)],
            'targets': [help_output_path],
            'file_dep': [output_path / help_file],
            'uptodate': [config_changed(page_flags)],
        }



from typing import Dict, List, Iterator, Optional, Tuple
from pathlib import Path
import re
import fnmatch
import itertools
import json

import attr

from elm_doc import elm_platform


ModuleName = str
ExactVersion = str
VersionRange = str
STUFF_DIRECTORY = 'elm-stuff'
module_name_re = re.compile(r'^[A-Z][a-zA-Z0-9_]*$')


@attr.s
class ElmProject:
    DOCS_FILENAME = 'docs.json'

    path = attr.ib()

    @property
    def json_path(self) -> Path:
        return self.path / self.DESCRIPTION_FILENAME

    def iter_dependencies(self) -> Iterator['ElmPackage']:
        raise NotImplementedError


@attr.s
class ElmPackage(ElmProject):
    DESCRIPTION_FILENAME = 'elm.json'
    PACKAGES_DIRECTORY = 'packages'

    user = attr.ib()  # str
    project = attr.ib()  # str
    version = attr.ib()  # str
    summary = attr.ib()  # str
    license = attr.ib()  # str
    elm_version = attr.ib()  # VersionRange
    exposed_modules = attr.ib()  # Union[List[ModuleName], Dict[str, List[ModuleName]]]
    dependencies = attr.ib()  # Dict[str, VersionRange]
    test_dependencies = attr.ib()  # Dict[str, VersionRange]

    @classmethod
    def from_path(cls, path: Path) -> Optional['ElmPackage']:
        json_path = path / cls.DESCRIPTION_FILENAME
        if not json_path.exists():
            return

        description = _load_json(json_path)
        if description['type'] != 'package':
            return

        name_parts = description['name'].split('/')
        return cls(
            path=path,
            user=name_parts[0],
            project=name_parts[1],
            version=description['version'],
            summary=description['summary'],
            license=description['license'],
            exposed_modules=description['exposed-modules'],
            dependencies=description['dependencies'],
            test_dependencies=description['test-dependencies'],
            elm_version=description['elm-version'],
        )

    @property
    def name(self) -> str:
        return '{}/{}'.format(self.user, self.project)

    def as_package(self, config):
        return self

    def as_json(self):
        fields = [
            ('name', 'name'),
            ('version', 'version'),
            ('summary', 'summary'),
            ('license', 'license'),
            ('exposed-modules', 'exposed_modules'),
            ('dependencies', 'dependencies'),
            ('test-dependencies', 'test_dependencies'),
            ('elm-version', 'elm_version'),
        ]
        props = {json_prop: getattr(self, attr) for json_prop, attr in fields}
        props['type'] = 'package'
        return props

    def without_license(self) -> 'ElmPackage':
        asdict = attr.asdict(self)
        asdict['license'] = ''
        return ElmPackage(**asdict)

    def sorted_exposed_modules(self):
        if isinstance(self.exposed_modules, dict):
            modules = [module for modules in self.exposed_modules.values()
                       for module in modules]
        else:
            modules = list(self.exposed_modules)
        return sorted(modules)


@attr.s
class ElmApplication(ElmProject):
    DESCRIPTION_FILENAME = 'elm.json'
    PACKAGES_DIRECTORY = 'package'

    source_directories = attr.ib()  # [str]
    elm_version = attr.ib()  # ExactVersion
    direct_dependencies = attr.ib()  # Dict[str, ExactVersion]
    indirect_dependencies = attr.ib()  # Dict[str, ExactVersion]
    direct_test_dependencies = attr.ib()  # Dict[str, ExactVersion]
    indirect_test_dependencies = attr.ib()  # Dict[str, ExactVersion]

    @classmethod
    def from_path(cls, path: Path) -> Optional['ElmApplication']:
        json_path = path / cls.DESCRIPTION_FILENAME
        if not json_path.exists():
            return

        description = _load_json(json_path)
        if description['type'] != 'application':
            return

        return cls(
            path=path,
            source_directories=description['source-directories'],
            elm_version=description['elm-version'],
            direct_dependencies=description['dependencies'].get('direct', {}),
            indirect_dependencies=description['dependencies'].get('indirect', {}),
            direct_test_dependencies=description['test-dependencies'].get('direct', {}),
            indirect_test_dependencies=description['test-dependencies'].get('indirect', {}),
        )

    def as_package(self, overrides: 'ProjectConfig') -> 'ElmPackage':
        return ElmPackage(
            path=self.path,
            user=overrides.fake_user,
            project=overrides.fake_project,
            version=overrides.fake_version,
            summary=overrides.fake_summary,
            license=overrides.fake_license,
            exposed_modules=[],
            dependencies=_as_package_dependencies(
                self.direct_dependencies, self.indirect_dependencies),
            test_dependencies=_as_package_dependencies(
                self.direct_test_dependencies, self.indirect_test_dependencies),
            elm_version=_as_version_range(self.elm_version),
        )

    def as_json(self):
        json = {
            'type': 'application',
            'source-directories': self.source_directories,
            'elm-version': self.elm_version,
        }

        json['dependencies'] = {}
        json['dependencies']['direct'] = self.direct_dependencies
        json['dependencies']['indirect'] = self.indirect_dependencies

        json['test-dependencies'] = {}
        json['test-dependencies']['direct'] = self.direct_test_dependencies
        json['test-dependencies']['indirect'] = self.indirect_test_dependencies

        return json

    def add_direct_dependencies(self, deps: List[Tuple[str, ExactVersion]]):
        for name, version in deps:
            self.direct_dependencies[name] = version

    def all_dependency_names(self) -> Iterator[str]:
        return itertools.chain(
            self.direct_dependencies.keys(),
            self.indirect_dependencies.keys(),
            self.direct_test_dependencies.keys(),
            self.indirect_test_dependencies.keys(),
        )

    def iter_dependencies(self) -> Iterator[ElmPackage]:
        deps = itertools.chain(
            self.direct_dependencies.items(),
            self.indirect_dependencies.items(),
            self.direct_test_dependencies.items(),
            self.indirect_test_dependencies.items(),
        )
        for name, version in deps:
            # e.g. ~/.elm/0.19.0/package/elm/core/1.0.0
            yield from_path(elm_platform.ELM_HOME / self.elm_version / self.PACKAGES_DIRECTORY / name / version)


def _as_package_dependencies(*app_dependencies: Dict[str, ExactVersion]) -> Dict[str, VersionRange]:
    package_deps = {}
    for app_deps in app_dependencies:
        for package_name, exact_version in app_deps.items():
            package_deps[package_name] = _as_version_range(exact_version)
    return package_deps


def _as_version_range(exact_version: ExactVersion) -> VersionRange:
    major, minor, patch = exact_version.split('.')
    next_version = '{}.{}.{}'.format(major, minor, int(patch, 10) + 1)
    return '{} <= v < {}'.format(exact_version, next_version)


def from_path(path: Path) -> ElmProject:
    classes = [
        ElmApplication,
        ElmPackage,
    ]
    for cls in classes:
        project = cls.from_path(path)
        if project:
            return project
    raise Exception('{} does not look like an Elm project'.format(path))


def _load_json(path: Path) -> Dict:
    with open(str(path)) as f:
        return json.load(f)


@attr.s
class ProjectConfig:
    include_paths = attr.ib(factory=list)  # List[str]
    exclude_modules = attr.ib(factory=list)  # List[str]
    force_exclusion = attr.ib(default=False)  # bool
    fake_user = attr.ib(default='user')  # str
    fake_project = attr.ib(default='project')  # str
    fake_version = attr.ib(default='1.0.0')  # str
    fake_summary = attr.ib(default='summary')  # str
    fake_license = attr.ib(default='BSD-3-Clause')  # str


@attr.s
class ElmModule:
    path = attr.ib()  # Path
    name = attr.ib()  # ModuleName


def glob_project_modules(
        project: ElmProject, config: ProjectConfig) -> Iterator[ElmModule]:
    for source_dir_name in project.source_directories:
        source_dir = project.path / source_dir_name
        elm_files = source_dir.glob('**/*.elm')
        for elm_file in elm_files:
            if elm_file.relative_to(project.path).parts[0] == STUFF_DIRECTORY:
                continue

            if config.include_paths and not any(_matches_include_path(elm_file, include_path)
                                                for include_path in config.include_paths):
                continue

            rel_path = elm_file.relative_to(source_dir)
            module_name_parts = rel_path.parent.parts + (rel_path.stem,)
            if not all(map(_valid_module_name, module_name_parts)):
                continue

            module_name = '.'.join(module_name_parts)

            # check for excludes if there's no explicit includes, or if
            # there are explicit includes and exclusion is requested specifically.
            check_excludes = (not config.include_paths) or config.force_exclusion
            if check_excludes and any(fnmatch.fnmatch(module_name, module_pattern)
                                      for module_pattern in config.exclude_modules):
                continue

            yield ElmModule(path=elm_file, name=module_name)


def _valid_module_name(name):
    return module_name_re.match(name)


def _matches_include_path(source_path: Path, include_path: Path):
    try:
        source_path.relative_to(include_path)
    except ValueError:
        return False
    else:
        return True



'''
note: this is by no means a complete Elm syntax parser. In fact,
it can only do little more than splitting a port declaration
at meaningful boundaries.
'''
from elm_doc.elm_parser.lines import ChunkType, Chunk, iter_line_chunks
from elm_doc.elm_parser.ports import PortInfo, parse_port_declaration

__all__ = [
    'ChunkType',
    'Chunk',
    'PortInfo',
    'iter_line_chunks',
    'parse_port_declaration',
]



from typing import Dict
import sys
import os.path
import io
import contextlib
import logging
from tempfile import TemporaryDirectory
import tarfile
import gzip
import hashlib
import json
import subprocess
import shutil
from pathlib import Path

sys.path.insert(0, str(Path(__file__).parent / 'src'))

from elm_doc import elm_platform, elm_project, tasks
from tests import conftest


logger = logging.getLogger(__name__)


def task_create_elm_core_fixture():
    workspace_path = Path(__file__).parent / 'workspace'
    elm_core_fixture_path = conftest.elm_core_fixture_path()
    elm_versions = ['0.19.0']
    for elm_version in elm_versions:
        build_tarball_path = Path(__file__).parent / 'build' / 'elm-core-fixture-{}.tar.gz'.format(elm_version)
        dist_tarball_path = Path(str(elm_core_fixture_path(elm_version)))
        yield {
            'basename': 'create_elm_core_fixture',
            'name': elm_version,
            'actions': [
                (_create_elm_core_fixture, (elm_version, build_tarball_path)),
                (_copy_if_tarball_changed, (build_tarball_path, dist_tarball_path)),
            ],
            'targets': [dist_tarball_path],
            'file_dep': [workspace_path / 'elm.json'],
        }


def _create_elm_core_fixture(elm_version: str, tarball: str):
    tarball.parent.mkdir(parents=True, exist_ok=True)
    workspace_path = Path(__file__).parent / 'workspace'
    # clear elm-stuff because something in there causes CORRUPT BINARY
    # if you run 'elm make' with a different ELM_HOME
    shutil.rmtree(str(workspace_path / elm_project.STUFF_DIRECTORY), ignore_errors=True)
    with TemporaryDirectory() as tmpdir:
        root_path = Path(tmpdir)
        elm_platform.install(root_path, elm_version)
        elm_path = root_path / 'node_modules' / '.bin' / 'elm'
        elm_home_path = root_path / '.elm'
        try:
            subprocess.check_output(
                [str(elm_path), 'make', str(workspace_path / 'src' / 'Main.elm')],
                env=dict(os.environ, **{'ELM_HOME': str(elm_home_path)}),
                cwd=str(workspace_path),
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            print('\nSTDOUT:\n' + e.stdout.decode('utf8'))
            print('\nSTDERR:\n' + e.stderr.decode('utf8'))
            raise e

        with _create_tarball(tarball) as tar:
            tar.add(str(elm_home_path), arcname=elm_home_path.name, filter=_tar_filter_without_registry)


def _tar_filter_without_registry(tarinfo):
    if Path(tarinfo.name).name == 'versions.dat':
        return
    return tarinfo


def task_create_package_elm_lang_org_artifact_tarball():
    build_tarball_path = Path(__file__).parent.joinpath('build', 'assets.tar.gz')
    dist_tarball_path = Path(__file__).parent.joinpath('src', 'elm_doc', 'assets', 'assets.tar.gz')
    return {
        'file_dep': [
            'build/package.elm-lang.org/artifacts/elm.js',
            'vendor/package.elm-lang.org/assets/favicon.ico',
            'vendor/package.elm-lang.org/assets/highlight/highlight.pack.js',
            'vendor/package.elm-lang.org/assets/highlight/LICENSE',
            'vendor/package.elm-lang.org/assets/highlight/styles/default.css',
            'vendor/package.elm-lang.org/assets/style.css',
            'vendor/package.elm-lang.org/assets/help/documentation-format.md',
            'vendor/package.elm-lang.org/assets/help/design-guidelines.md',
            'vendor/package.elm-lang.org/LICENSE',
        ],
        'targets': [dist_tarball_path],
        'actions': [
            (_create_package_elm_lang_org_artifact_tarball, (build_tarball_path,)),
            (_copy_if_tarball_changed, (build_tarball_path, dist_tarball_path)),
        ],
    }


def _copy_if_tarball_changed(source_path: Path, target_path: Path):
    if _is_tarball_different(source_path, target_path):
        shutil.copyfile(str(source_path), str(target_path))


def _is_tarball_different(source_path: Path, target_path: Path) -> bool:
    if not target_path.exists():
        return True
    source_md5s = _get_tarball_md5s(source_path)
    target_md5s = _get_tarball_md5s(target_path)
    if source_md5s != target_md5s:
        for name in (set(source_md5s.keys()) | set(target_md5s.keys())):
            if source_md5s.get(name) != target_md5s.get(name):
                logger.info('hash of {} differ between {} and {}'.format(name, source_path, target_path))
        return True
    return False


def _get_tarball_md5s(path: Path) -> Dict[str, str]:
    res = {}
    with tarfile.open(str(path), 'r') as tar:
        for member in tar.getmembers():
            if not member.isfile():
                continue
            f = tar.extractfile(member)
            md5 = hashlib.md5()
            block_size = 128 * md5.block_size
            while True:
                data = f.read(block_size)
                if not data:
                    break
                md5.update(data)
            res[member.name] = md5.hexdigest()
    return res


def _create_package_elm_lang_org_artifact_tarball(output_path: Path):
    build_artifacts_path = Path(__file__).parent.joinpath('build', 'package.elm-lang.org', 'artifacts')
    vendor_assets_path = Path(__file__).parent.joinpath('vendor', 'package.elm-lang.org', 'assets')
    vendor_license_path = Path(__file__).parent.joinpath('vendor', 'package.elm-lang.org', 'LICENSE')
    with _create_tarball(output_path) as tar:
        tar.add(str(build_artifacts_path), arcname=build_artifacts_path.name)
        tar.add(str(vendor_assets_path), arcname=vendor_assets_path.name)
        tar.add(str(vendor_license_path), arcname='assets/LICENSE')
        tar.add(str(vendor_license_path), arcname='artifacts/LICENSE')


@contextlib.contextmanager
def _create_tarball(output_path: Path):
    output_path.parent.mkdir(parents=True, exist_ok=True)
    tar_bytes = io.BytesIO()
    with tarfile.open(fileobj=tar_bytes, mode="w") as tar:
        yield tar
    tar_bytes.seek(0)
    # Hardcode filename and mtime so that the output is deterministic
    # as much as possible.  This is equivalent to doing `tar .. | gzip
    # -n` on the command line.  We *could* strip mtime from the
    # tarball entries as well, but having that information included
    # may be useful for debugging in the future, so we avoid updating
    # the tarball by checking the md5 of the contents above instead.
    with open(str(output_path), 'wb') as f:
        with gzip.GzipFile(filename='', mode='wb', fileobj=f, mtime=0) as z:
            z.write(tar_bytes.read())


def task_package_elm_lang_org_elm_js():
    root_path = Path(__file__).parent.joinpath('vendor', 'package.elm-lang.org')
    output_path = Path(__file__).parent.joinpath('build', 'package.elm-lang.org', 'artifacts', 'elm.js')
    return {
        'file_dep': list(root_path.joinpath('src/frontend').glob('**/*.elm')) + [root_path / 'elm.json'],
        'targets': [
            'build/package.elm-lang.org/artifacts/elm.js',
        ],
        'actions': [(_create_package_elm_lang_org_elm_js, (output_path,))]
    }


def _read_elm_version(elm_json_path: Path) -> str:
    elm_json = json.loads(elm_json_path.read_text('utf8'))
    return elm_json['elm-version']


def _create_package_elm_lang_org_elm_js(output_path: Path):
    repo_path = Path(__file__).parent / 'vendor' / 'package.elm-lang.org'
    elm_version = _read_elm_version(repo_path / 'elm.json')
    with TemporaryDirectory() as tmpdir:
        root_path = Path(tmpdir)
        elm_platform.install(root_path, elm_version)
        elm_path = root_path / 'node_modules' / '.bin' / 'elm'
        try:
            subprocess.check_output(
                [
                    str(elm_path),
                    'make',
                    str(repo_path / 'src' / 'frontend' / 'Main.elm'),
                    '--output',
                    str(output_path),
                ],
                cwd=str(repo_path),
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            print('\nSTDOUT:\n' + e.stdout.decode('utf8'))
            print('\nSTDERR:\n' + e.stderr.decode('utf8'))
            raise e

# build docs for the workspace project

workspace_path = Path(__file__).parent / 'workspace'
output_path = workspace_path / 'build' / 'docs'
elm_path = workspace_path / 'node_modules' / '.bin' / 'elm'
config = elm_project.ProjectConfig()


def task_install_workspace_elm():
    yield {
        'name': 'install_elm',
        'file_dep': [workspace_path / 'elm.json'],
        'targets': [elm_path],
        'actions': [(_install_elm, (workspace_path,))]
    }


def _install_elm(project_path: Path):
    elm_version = _read_elm_version(project_path / 'elm.json')
    elm_platform.install(project_path, elm_version)


for creator_name, creator_func in tasks.build_task_creators(
        workspace_path,
        config,
        output_path,
        elm_path=elm_path,
        mount_point='/docs',
).items():
    globals()[creator_name] = creator_func



from typing import List, Optional
from collections import ChainMap
from pathlib import Path
import json
from tempfile import TemporaryDirectory
import subprocess

from dirsync import sync
from doit.tools import create_folder, config_changed

from elm_doc import elm_platform
from elm_doc import elm_project
from elm_doc import elm_codeshift
from elm_doc import package_tasks
from elm_doc.elm_project import ElmPackage, ElmProject, ProjectConfig, ModuleName
from elm_doc.decorators import capture_subprocess_error


@capture_subprocess_error
def build_project_docs_json(
        project: ElmProject,
        project_config: ProjectConfig,
        project_modules: List[ModuleName],
        output_path: Path = None,
        build_path: Path = None,
        elm_path: Path = None,
        validate: bool = False):
    elm_project_with_exposed_modules = dict(ChainMap(
        {'exposed-modules': [module for module in project_modules]},
        project.as_package(project_config).as_json(),
    ))
    with TemporaryDirectory() as tmpdir:
        tmp_path = Path(tmpdir)

        build_path.mkdir(parents=True, exist_ok=True)

        elm_json_path = build_path / ElmPackage.DESCRIPTION_FILENAME
        with open(str(elm_json_path), 'w') as f:
            json.dump(elm_project_with_exposed_modules, f)

        package_src_dir = build_path / 'src'
        changed_files = set()
        for source_dir in project.source_directories:
            changed_files |= sync(
                str(project.path / source_dir), str(package_src_dir), 'sync', create=True)

        for elm_file in changed_files:
            elm_file_path = Path(elm_file)
            if elm_file_path.suffix == '.elm' and elm_file_path.exists():
                elm_codeshift.strip_ports_from_file(Path(elm_file))

        if elm_path is None:
            elm_platform.install(tmp_path, project.elm_version)
            elm_path = elm_platform.get_node_modules_elm_path(tmp_path)

        if validate:
            # don't update the final artifact; write to build dir instead
            output_path = build_path / project.DOCS_FILENAME

        subprocess.check_output(
            [str(elm_path), 'make', '--docs', str(output_path), '--output', '/dev/null'],
            cwd=str(build_path),
            stderr=subprocess.STDOUT,
        )


def create_main_project_tasks(
        project: ElmProject,
        project_config: ProjectConfig,
        output_path: Optional[Path],
        build_path: Path = None,
        elm_path: Path = None,
        mount_point: str = '',
        validate: bool = False):
    task_name = '{}/{}'.format(project_config.fake_user, project_config.fake_project)
    project_modules = list(elm_project.glob_project_modules(
        project, project_config))
    project_as_package = project.as_package(project_config)
    file_dep = [elm_path] if elm_path else []
    file_dep.extend([module.path for module in project_modules])
    uptodate_config = {'elm_json': project_as_package.as_json()}

    if validate:
        yield {
            'basename': 'validate_docs_json',
            'name': task_name,
            'actions': [(build_project_docs_json,
                         (project, project_config, [module.name for module in project_modules]),
                         {'build_path': build_path, 'elm_path': elm_path, 'validate': True})],
            'targets': [],
            'file_dep': file_dep,
            'uptodate': [config_changed(uptodate_config)],
        }
        return

    project_output_path = package_tasks.package_docs_root(output_path, project_as_package)

    # project docs.json
    docs_json_path = project_output_path / project.DOCS_FILENAME
    yield {
        'basename': 'build_docs_json',
        'name': task_name,
        'actions': [(create_folder, (str(project_output_path),)),
                    (build_project_docs_json,
                     (project, project_config, [module.name for module in project_modules]),
                     {'build_path': build_path, 'elm_path': elm_path, 'output_path': docs_json_path})],
        'targets': [docs_json_path],
        'file_dep': file_dep,
        'uptodate': [config_changed(uptodate_config)],
    }

    for page_task in package_tasks.create_package_page_tasks(
            package_tasks.Context.Project,
            output_path,
            project_as_package,
            [module.name for module in project_modules],
            mount_point):
        yield page_task



import functools
import subprocess

from doit.exceptions import TaskFailed


def capture_subprocess_error(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except subprocess.CalledProcessError as e:
            command_string = e.cmd if isinstance(e.cmd, str) else ' '.join(e.cmd)
            return TaskFailed(
                'Error while executing {}:\nstdout:\n{}\n\nstderr:\n{}'.format(
                    command_string,
                    e.stdout.decode('utf8') if e.stdout else '',
                    e.stderr.decode('utf8') if e.stderr else ''))
    return wrapper



import os
import subprocess
from pathlib import Path
import json

from elm_doc import node_modules
from elm_doc.decorators import capture_subprocess_error


# See: https://github.com/elm/compiler/blob/0.19.0/builder/src/Elm/PerUserCache.hs#L44
# Note: this implementation is not exactly the same as the above, which uses
# getAppUserDataDirectory:
# http://hackage.haskell.org/package/directory-1.3.3.1/docs/System-Directory.html#v:getAppUserDataDirectory
# On Windows, this is something like 'C:/Users/<user>/AppData/Roaming/<app>'.
# elm-doc currently don't support Windows, so this is fine for now.
ELM_HOME = Path(os.environ['ELM_HOME']) if 'ELM_HOME' in os.environ else (Path.home() / '.elm')


def install(to: Path, elm_version: str):
    npm_package = {
        'dependencies': {
            'elm': get_npm_version_range(elm_version)
        }
    }
    with open(str(to / 'package.json'), 'w') as f:
        json.dump(npm_package, f)
    node_modules.install(cwd=str(to))


@capture_subprocess_error
def get_node_modules_elm_path(project_root: Path):
    script = 'console.log(require.resolve("elm"))'
    # e.g. path/to/node_modules/elm/index.js
    elm_index_path = subprocess.check_output(
        ['node', '-e', script],
        cwd=str(project_root),
    )
    # e.g. path/to/node_modules/elm/unpacked_bin/elm
    return Path(elm_index_path.decode('utf-8').strip()).parent / 'unpacked_bin' / 'elm'


def get_npm_version_range(elm_version: str) -> str:
    if _is_exact(elm_version):
        return elm_version
    min_version, gt_op, _, lt_op, max_version = elm_version.split(' ')
    return '{gt_op}{min_version} {lt_op}{max_version}'.format(
        min_version=min_version,
        gt_op=_flip_inequality_op(gt_op),
        lt_op=lt_op,
        max_version=max_version,
    )


def _is_exact(elm_version: str) -> bool:
    return ' ' not in elm_version


def _flip_inequality_op(op: str) -> str:
    # assume there's only one < or >
    return op.replace('>', '<').replace('<', '>')



from typing import List
import logging
from pathlib import Path
import itertools

import parsy

from elm_doc import elm_parser
from elm_doc.elm_parser import Chunk


logger = logging.getLogger(__name__)


def strip_ports_from_file(elm_file: Path) -> None:
    source = elm_file.read_text()
    stripped = strip_ports_from_string(source, str(elm_file))
    elm_file.write_text(stripped)


def strip_ports_from_string(source: str, source_path: str = '<unknown>') -> str:
    output = []
    for chunk in elm_parser.iter_line_chunks(source):
        if chunk.is_port_function():
            output.extend(_rewrite_port_function(chunk, source_path))
        elif chunk.is_port_module():
            output.extend(_rewrite_port_module(chunk))
        else:
            output.extend(chunk.raw_lines())
    return ''.join(output)


def _rewrite_port_module(chunk: Chunk) -> List[str]:
    output = []
    for line in chunk.lines:
        if line.raw.startswith('port module'):
            output.append(line.raw[len('port '):])
        else:
            output.append(line.raw)
    return output


def _rewrite_port_function(chunk: Chunk, source_path: str) -> List[str]:
    output = []

    trailing_non_sources = list(reversed(list(itertools.takewhile(
        lambda l: not l.is_source(), reversed(chunk.lines)))))
    declaration_lines = chunk.lines[:-len(trailing_non_sources)] if trailing_non_sources else chunk.lines

    for line in declaration_lines:
        if line.raw.startswith('port '):
            output.append(line.raw[len('port '):])
        else:
            # These lines may contain comments which, semantically
            # speaking, belong to the next function: the output may
            # not make sense to human reader. Since our audience is
            # the Elm compiler, we don't care that much.
            output.append(line.raw)

    # If this chunk is the very last lines of a file, we may need to add a newline.
    if (not trailing_non_sources) and (not chunk.lines[-1].raw.endswith('\n')):
        output.append('\n')

    try:
        one_liner = ''.join([raw.strip() for raw in chunk.non_comment_raw_lines()])
        port_info = elm_parser.parse_port_declaration(one_liner)
        output.append(_make_dummy_port_implementation(port_info) + '\n')
    except parsy.ParseError:
        line_num = chunk.lines[0].number
        logger.error('''{}:{}: failed to parse a port declaration. We parsed it as:

  {}

If this is a valid port declaration, please submit a bug report'''.format(
            source_path, line_num, one_liner))
        raise

    for trailing_line in trailing_non_sources:
        output.append(trailing_line.raw)

    return output


def _make_dummy_port_implementation(port_info: elm_parser.PortInfo) -> str:
    parts = [port_info.name]

    # arguments
    for i, _ in enumerate(port_info.args):
        parts.append('a{}'.format(i))

    parts.append('=')
    parts.append('{}.none'.format(port_info.port_type.value))
    return ' '.join(parts)



from typing import List, Optional
import enum
from pathlib import Path
import json
import shutil

from doit.tools import create_folder, config_changed

from elm_doc import page_tasks
from elm_doc.elm_project import ElmPackage, ModuleName


def build_package_releases(output_path: Path, version: str, timestamp: int):
    releases = {version: timestamp}
    with open(str(output_path), 'w') as f:
        json.dump(releases, f)


def link_latest_package_dir(output_path: Path, package_dir: Path):
    package_dir.mkdir(parents=True, exist_ok=True)
    # prefer relative path to make the built documentation directory relocatable
    if output_path.is_symlink():
        output_path.unlink()
    output_path.symlink_to(package_dir.relative_to(output_path.parent), target_is_directory=True)


def copy_package_readme(package_readme: Path, output_path: Path):
    if package_readme.is_file():
        shutil.copy(str(package_readme), str(output_path))


def copy_package_docs_json(package: ElmPackage, output_path: Path):
    source = package.path / package.DOCS_FILENAME
    shutil.copy(str(source), str(output_path))


def _package_task_name(package):
    return '{}/{}'.format(package.name, package.version)


def create_dependency_tasks(
        output_path: Optional[Path],
        package: ElmPackage,
        mount_point: str = ''):
    task_name = _package_task_name(package)
    package_modules = package.sorted_exposed_modules()
    package_output_path = package_docs_root(output_path, package)

    # package docs.json
    docs_json_path = package_output_path / package.DOCS_FILENAME
    yield {
        'basename': Context.Dependency.basename('copy_docs_json'),
        'name': task_name,
        'actions': [(create_folder, (str(package_output_path),)),
                    (copy_package_docs_json, (package, docs_json_path))],
        'targets': [docs_json_path],
        'file_dep': [package.path / package.DOCS_FILENAME]
    }

    for page_task in create_package_page_tasks(Context.Dependency, output_path, package, package_modules, mount_point):
        yield page_task


class Context(enum.Enum):
    '''In order to delay the creation of tasks for project dependencies,
    the basenames of the delayed tasks need to be different than those
    for the main project tasks for doit to correctly pick up.

    This enum provides a way for differentiating those two contexts when
    creating tasks in this module, which are shared between the two.
    '''
    Project = 'project'
    Dependency = 'dep'

    def basename(self, suffix):
        return '{}_{}'.format(self.value, suffix)


def create_package_page_tasks(
        context: Context,
        output_path: Optional[Path],
        package: ElmPackage,
        package_modules: List[ModuleName],
        mount_point: str = '',):
    task_name = _package_task_name(package)
    package_output_path = package_docs_root(output_path, package)
    page_flags = {'mount_point': mount_point}

    # package index page
    package_index_output = package_output_path / 'index.html'
    yield {
        'basename': context.basename('top_page'),
        'name': task_name,
        'actions': [(page_tasks.write_page, (package_index_output,), page_flags)],
        'targets': [package_index_output],
        'uptodate': [config_changed(page_flags)],
    }

    # package readme
    readme_filename = 'README.md'
    package_readme = package.path / readme_filename
    output_readme_path = package_output_path / readme_filename
    if package_readme.is_file():
        yield {
            'basename': context.basename('readme'),
            'name': task_name,
            'actions': [(copy_package_readme, (package_readme, output_readme_path))],
            'targets': [output_readme_path],
            'file_dep': [package_readme],
        }

    # package releases
    package_releases_output = package_output_path.parent / 'releases.json'
    # generated docs will only ever include one version per package, so we hardcode
    # the timestamp.
    content = {'version': package.version, 'timestamp': 1}
    yield {
        'basename': context.basename('releases'),
        'name': task_name,
        'actions': [(build_package_releases, (package_releases_output,), content)],
        'targets': [package_releases_output],
        'uptodate': [config_changed(content)]
    }

    # link from /latest
    latest_path = package_output_path.parent / 'latest'
    uptodate_config = {'link_target': str(package_output_path.relative_to(output_path))}
    yield {
        'basename': context.basename('latest_link'),
        'name': task_name,
        'actions': [(link_latest_package_dir, (latest_path, package_output_path))],
        'targets': [latest_path],
        'uptodate': [config_changed(uptodate_config)]
    }

    # module pages
    for module_name in package_modules:
        module_output = package_output_path / module_name.replace('.', '-')
        yield {
            'basename': context.basename('module_page'),
            'name': '{}:{}'.format(task_name, module_name),
            'actions': [(page_tasks.write_page, (module_output,), page_flags)],
            'targets': [module_output],
            'uptodate': [config_changed(page_flags)],
        }


def package_docs_root(output_path: Optional[Path], package: ElmPackage) -> Path:
    return output_path / 'packages' / package.user / package.project / package.version



from pathlib import Path
import tarfile


tarball = Path(__file__).parent / 'assets' / 'assets.tar.gz'
bundled_helps = [
    'assets/help/documentation-format.md',
    'assets/help/design-guidelines.md',
]
bundled_assets = bundled_helps + [
    'artifacts/elm.js',
    'artifacts/LICENSE',
    'assets/favicon.ico',
    'assets/highlight/highlight.pack.js',
    'assets/highlight/LICENSE',
    'assets/highlight/styles/default.css',
    'assets/LICENSE',
    'assets/style.css',
]


def extract_assets(output_path: Path):
    with tarfile.open(str(tarball)) as f:
        f.extractall(str(output_path))



import json
from pathlib import Path


PAGE_TEMPLATE = '''
<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <link rel="shortcut icon" size="16x16, 32x32, 48x48, 64x64, 128x128, 256x256" href="{mount_point}/assets/favicon.ico">
    <link rel="stylesheet" href="{mount_point}/assets/highlight/styles/default.css">
    <link rel="stylesheet" href="{mount_point}/assets/style.css">
    <script src="{mount_point}/assets/highlight/highlight.pack.js"></script>
    <script src="{mount_point}/artifacts/elm.js"></script>
  </head>
  <body>
  <script>
    Elm.Main.init({init});
  </script>
  </body>
</html>
'''  # noqa: E501


def _render(mount_point: str = ''):
    if mount_point and mount_point[-1] == '/':
        mount_point = mount_point[:-1]
    init = {
        'flags': {
            'mountedAt': mount_point,
        },
    }
    return PAGE_TEMPLATE.format(mount_point=mount_point, init=json.dumps(init))


def write_page(output_path: Path, mount_point: str = ''):
    output_path.parent.mkdir(parents=True, exist_ok=True)
    with open(str(output_path), 'w') as f:
        f.write(_render(mount_point=mount_point))



'''
Run npm commands. Uses yarn if present.
'''
import shutil
import subprocess

from elm_doc.decorators import capture_subprocess_error


def has_yarn():
    return shutil.which('yarn') is not None


@capture_subprocess_error
def install(*extra_args, **popen_kwargs):
    if has_yarn():
        subprocess.check_output(
            ('yarn', 'install') + extra_args,
            stderr=subprocess.STDOUT,
            **popen_kwargs)
    else:
        subprocess.check_output(
            ('npm', 'install') + extra_args,
            stderr=subprocess.STDOUT,
            **popen_kwargs)


@capture_subprocess_error
def add(*extra_args, **popen_kwargs):
    if has_yarn():
        subprocess.check_output(
            ('yarn', 'add') + extra_args,
            stderr=subprocess.STDOUT,
            **popen_kwargs)
    else:
        subprocess.check_output(
            ('npm', 'install') + extra_args,
            stderr=subprocess.STDOUT,
            **popen_kwargs)



"""
Generate static documentation for your Elm application.
"""
import sys
from setuptools import find_packages, setup


setup(
    name='elm-doc',
    version='1.0.0b1',
    url='https://github.com/ento/elm-doc',
    license='BSD 3-clause',
    author='Marica Odagaki',
    author_email='ento.entotto@gmail.com',
    description='Generate static documentation for your Elm application',
    long_description=__doc__,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'click',
        'dirsync',
        'doit',
        'parsy',
        'requests',
        'retrying',
    ],
    packages=find_packages('src', exclude=['tests']),
    package_dir={'': 'src'},
    package_data={'': ['elm_doc/assets/assets.tar.gz']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'elm-doc = elm_doc.cli:main',
        ],
    },
    python_requires='>=3.5',
    classifiers=[
        # As from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        # 'Development Status :: 5 - Production/Stable',
        # 'Development Status :: 6 - Mature',
        # 'Development Status :: 7 - Inactive',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: POSIX',
        'Operating System :: MacOS',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Software Development :: Documentation',
    ]
)



'''
Functions for parsing a port declaration.
'''
import enum

from parsy import generate, regex, string
import attr


class PortType(enum.Enum):
    Command = 'Cmd'
    Subscription = 'Sub'


@attr.s
class PortInfo:
    name = attr.ib()  # str
    args = attr.ib()  # [str]
    port_type = attr.ib()  # PortType


def lexeme(p):
    """
    From a parser (or string), make a parser that consumes
    whitespace on either side.
    """
    if isinstance(p, str):
        p = string(p)
    return regex(r'\s*') >> p << regex(r'\s*')


func_name = regex(r'[a-z][a-zA-Z0-9_]*').desc('function name')
no_parens_argument = lexeme(regex(r'[^({][^-]+')).desc('arg without outer parens')


@generate
def parens_argument():
    opening = yield lexeme(regex(r'[({]')).desc('opening brace')

    content = yield parens_content.many()

    parens_pairs = {'(': ')', '{': '}'}
    expected_closing = lexeme(parens_pairs[opening]).desc('closing brace')
    closing = yield expected_closing

    return opening + ''.join(content) + closing


string_without_parens = lexeme(regex(r'[^(){}]+')).desc('string w/o parens')
parens_content = lexeme('()') | string_without_parens | parens_argument
argument = lexeme('()') | no_parens_argument | parens_argument


@generate
def port_info():
    '''Parse port declaration. This parser only concerns itself with information
    necessary for producing a dummy implementation of the function in order to
    make the elm compiler happy when generating documentation for a project disguised
    as a package. Ports are not allowed in packages, so we need to transform them
    into ordinary functions.
    '''
    yield lexeme('port')
    port_name = yield lexeme(func_name)
    yield lexeme(':')
    port_args = yield (argument << lexeme('->')).many()
    port_value = yield lexeme('Cmd') | lexeme('Sub')
    port_type = PortType(port_value)
    yield regex(r'\S+').desc('remainder')
    return PortInfo(
        name=port_name,
        args=[arg.strip() for arg in port_args],
        port_type=port_type)


parse_port_declaration = port_info.parse
